import importlib
import sys
from dataclasses import dataclass
from logging import getLogger
from pathlib import Path
from typing import Tuple, Union

from rich import print
from rich.padding import Padding
from rich.panel import Panel
from rich.syntax import Syntax
from rich.tree import Tree

from .. import FastAgency
from ..exceptions import FastAgencyCLIError

logger = getLogger(__name__)


def get_default_path() -> Path:
    potential_paths = (
        "main.py",
        "app.py",
        "api.py",
        "app/main.py",
        "app/app.py",
        "app/api.py",
    )

    for full_path in potential_paths:
        path = Path(full_path)
        if path.is_file():
            return path

    raise FastAgencyCLIError(
        "Could not find a default file to run, please provide an explicit path"
    )


@dataclass
class ModuleData:
    module_import_str: str
    extra_sys_path: Path


def get_module_data_from_path(path: Path) -> ModuleData:
    logger.info(
        "Searching for package file structure from directories with [blue]__init__.py[/blue] files"
    )
    use_path = path.resolve()
    module_path = use_path
    if use_path.is_file() and use_path.stem == "__init__":
        module_path = use_path.parent
    module_paths = [module_path]
    extra_sys_path = module_path.parent
    for parent in module_path.parents:
        init_path = parent / "__init__.py"
        if init_path.is_file():
            module_paths.insert(0, parent)
            extra_sys_path = parent.parent
        else:
            break
    logger.info(f"Importing from {extra_sys_path.resolve()}")
    root = module_paths[0]
    name = f"🐍 {root.name}" if root.is_file() else f"📁 {root.name}"
    root_tree = Tree(name)
    if root.is_dir():
        root_tree.add("[dim]🐍 __init__.py[/dim]")
    tree = root_tree
    for sub_path in module_paths[1:]:
        sub_name = (
            f"🐍 {sub_path.name}" if sub_path.is_file() else f"📁 {sub_path.name}"
        )
        tree = tree.add(sub_name)
        if sub_path.is_dir():
            tree.add("[dim]🐍 __init__.py[/dim]")
    title = "[b green]Python module file[/b green]"
    if len(module_paths) > 1 or module_path.is_dir():
        title = "[b green]Python package file structure[/b green]"
    panel = Padding(
        Panel(
            root_tree,
            title=title,
            expand=False,
            padding=(1, 2),
        ),
        1,
    )
    print(panel)
    module_str = ".".join(p.stem for p in module_paths)
    logger.info(f"Importing module [green]{module_str}[/green]")
    return ModuleData(
        module_import_str=module_str,
        extra_sys_path=extra_sys_path.resolve(),
    )


def get_app_name(  # noqa: C901
    *, mod_data: ModuleData, app_name: Union[str, None] = None
) -> "Tuple[str, FastAgency]":
    try:
        mod = importlib.import_module(mod_data.module_import_str)  # nosemgrep
    except (ImportError, ValueError) as e:
        logger.error(f"Import error: {e}")
        logger.warning(
            "Ensure all the package directories have an [blue]__init__.py[/blue] file"
        )
        raise
    if not FastAgency:  # type: ignore[truthy-function]
        raise FastAgencyCLIError(
            "Could not import FastAgency, try running 'pip install fastagency'"
        ) from None
    object_names = dir(mod)
    object_names_set = set(object_names)
    if app_name:
        if app_name not in object_names_set:
            raise FastAgencyCLIError(
                f"Could not find app name {app_name} in {mod_data.module_import_str}"
            )
        app = getattr(mod, app_name)
        if not isinstance(app, FastAgency):
            raise FastAgencyCLIError(
                f"The app name {app_name} in {mod_data.module_import_str} doesn't seem to be a FastAgency app"
            )
        return app_name, app
    for preferred_name in ["app", "api"]:
        if preferred_name in object_names_set:
            obj = getattr(mod, preferred_name)
            if isinstance(obj, FastAgency):
                return preferred_name, obj
    for name in object_names:
        obj = getattr(mod, name)
        if isinstance(obj, FastAgency):
            return name, obj
    raise FastAgencyCLIError("Could not find FastAgency app in module, try using --app")


def get_import_string(
    *, path: Union[Path, None] = None, app_name: Union[str, None] = None
) -> tuple[str, FastAgency]:
    if not path:
        path = get_default_path()
    logger.info(f"Using path [blue]{path}[/blue]")
    logger.info(f"Resolved absolute path {path.resolve()}")
    if not path.exists():
        raise FastAgencyCLIError(f"Path does not exist {path}")
    mod_data = get_module_data_from_path(path)
    sys.path.insert(0, str(mod_data.extra_sys_path))
    use_app_name, app = get_app_name(mod_data=mod_data, app_name=app_name)
    import_example = Syntax(
        f"from {mod_data.module_import_str} import {use_app_name}", "python"
    )
    import_panel = Padding(
        Panel(
            import_example,
            title="[b green]Importable FastAgency app[/b green]",
            expand=False,
            padding=(1, 2),
        ),
        1,
    )
    logger.info("Found importable FastAgency app")
    print(import_panel)
    import_string = f"{mod_data.module_import_str}:{use_app_name}"
    logger.info(f"Using import string [b green]{import_string}[/b green]")
    return import_string, app


def import_from_string(import_string: str) -> FastAgency:
    """Import a module and attribute from an import string.

    Import a module and an attribute from a string like 'file_name:app_name'.
    Checks if the file exists before attempting to import the module.

    Args:
        import_string (str): The import string in 'module_name:attribute_name' format.

    Returns:
        Any: The attribute from the module.

    Raises:
        ImportError: If the import string is not in the correct format or the module or attribute is not found.
        ValueError: If the import string is not in 'module_name:attribute_name' format.
        ModuleNotFoundError: If the module is not found.
        AttributeError: If the attribute is not found in the module.

    """
    try:
        # Split the string into module and attribute parts
        module_name, attribute_name = import_string.split(":")

        # Ensure the module name points to a valid Python file before importing
        module_path = f"{module_name.replace('.', '/')}.py"
        if not Path(module_path).is_file():
            raise ImportError(f"The file for module '{module_name}' does not exist.")

        # Add the current directory to the Python path to allow imports from local files
        sys.path.append(str(Path.cwd()))

        # Import the module using importlib
        module = importlib.import_module(module_name)  # nosemgrep

        # Get the attribute (like 'app') from the module
        attribute = getattr(module, attribute_name)
        if not isinstance(attribute, FastAgency):
            raise ImportError(
                f"The attribute '{attribute_name}' in module '{module_name}' is not a FastAgency app."
            )

        return attribute
    except ValueError:
        raise ImportError(
            "Import string must be in 'module_name:attribute_name' format."
        ) from None
    except ModuleNotFoundError:
        raise ImportError(f"Module '{module_name}' not found.") from None
    except AttributeError:
        raise ImportError(
            f"Attribute '{attribute_name}' not found in module '{module_name}'."
        ) from None



import json
import re
from functools import cached_property, wraps
from typing import Any

import stringcase
from fastapi_code_generator import __main__ as fastapi_code_generator_main
from fastapi_code_generator.parser import OpenAPIParser, Operation

from ...logging import get_logger

logger = get_logger(__name__)


def patch_parse_schema() -> None:
    org_parse_schema = OpenAPIParser.parse_schema

    @wraps(org_parse_schema)
    def my_parse_schema(*args: Any, **kwargs: Any) -> Any:
        data_type = org_parse_schema(*args, **kwargs)
        if data_type.reference and data_type.reference.duplicate_name:
            data_type.reference.name = data_type.reference.duplicate_name
        return data_type

    OpenAPIParser.parse_schema = my_parse_schema
    logger.info("Patched OpenAPIParser.parse_schema")


def patch_function_name_parsing() -> None:
    def function_name(self: Operation) -> str:
        if self.operationId:
            name: str = self.operationId.replace("/", "_")
        else:
            path = re.sub(r"/{|/", "_", self.snake_case_path).replace("}", "")
            name = f"{self.type}{path}"
        return stringcase.snakecase(name)  # type: ignore[no-any-return]

    Operation.function_name = cached_property(function_name)
    Operation.function_name.__set_name__(Operation, "function_name")

    logger.info("Patched Operation.function_name")


def patch_generate_code() -> None:
    # Save reference to the original generate_code function
    org_generate_code = fastapi_code_generator_main.generate_code

    @wraps(org_generate_code)
    def patched_generate_code(*args: Any, **kwargs: Any) -> Any:
        try:
            input_text: str = kwargs["input_text"]

            json_spec = json.loads(input_text)

            schemas_with_dots = sorted(
                [
                    name
                    for name in json_spec.get("components", {}).get("schemas", {})
                    if "." in name
                ],
                key=len,
                reverse=True,  # Sort by length in descending order
            )

            for schema_name in schemas_with_dots:
                new_schema_name = schema_name.replace(".", "_")
                input_text = input_text.replace(schema_name, new_schema_name)

            kwargs["input_text"] = input_text

        except Exception as e:
            logger.info(
                f"Patched fastapi_code_generator.__main__.generate_code raised: {e}, passing untouched arguments to original generate_code"
            )

        return org_generate_code(*args, **kwargs)

    fastapi_code_generator_main.generate_code = patched_generate_code

    logger.info("Patched fastapi_code_generator.__main__.generate_code")



import re
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field, fields
from typing import Any, Literal, Optional, Protocol, Type
from uuid import UUID, uuid4

from pydantic import BaseModel

from .logging import get_logger

__all__ = [
    "IOMessage",
    "TextMessage",
    "SuggestedFunctionCall",
    "FunctionCallExecution",
    "TextInput",
    "MultipleChoice",
    "SystemMessage",
    "KeepAlive",
    "WorkflowStarted",
    "WorkflowCompleted",
    "Error",
    "MessageType",
    "MessageProcessorProtocol",
    "MessageProcessorMixin",
]


logger = get_logger(__name__)


def _camel_to_snake(name: str) -> str:
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


# we keep this hardcoded for mypy and type checks
MessageType = Literal[
    "text_message",
    "suggested_function_call",
    "function_call_execution",
    "text_input",
    "multiple_choice",
    "system_message",
    "keep_alive",
    "workflow_started",
    "workflow_completed",
    "error",
]


@dataclass
class IOMessage(
    ABC
):  # `IOMessage` is an abstract base class, but it has no abstract methods
    workflow_uuid: str
    sender: Optional[str] = None
    recipient: Optional[str] = None
    auto_reply: bool = False
    uuid: str = field(default_factory=lambda: str(uuid4().hex))

    @property
    def type(self) -> MessageType:
        retval: MessageType = _camel_to_snake(self.__class__.__name__)  # type: ignore[assignment]
        return retval

    @staticmethod
    def _get_message_class(type: Optional[MessageType]) -> "Type[IOMessage]":
        type = type or "text_message"
        lookup: dict[MessageType, "Type[IOMessage]"] = {
            "text_message": TextMessage,
            "suggested_function_call": SuggestedFunctionCall,
            "function_call_execution": FunctionCallExecution,
            "text_input": TextInput,
            "multiple_choice": MultipleChoice,
            "keep_alive": KeepAlive,
            "system_message": SystemMessage,
            "workflow_started": WorkflowStarted,
            "workflow_completed": WorkflowCompleted,
            "error": Error,
        }
        return lookup[type]

    @staticmethod
    def create(type: Optional[MessageType] = None, **kwargs: Any) -> "IOMessage":
        cls = IOMessage._get_message_class(type)

        content = kwargs.pop("content", {})
        kwargs.update(content)

        return cls(**kwargs)

    @staticmethod
    def _get_parameters_names() -> list[str]:
        return [field.name for field in fields(IOMessage)]

    def model_dump(self) -> dict[str, Any]:
        params_names = IOMessage._get_parameters_names()
        d = asdict(self)
        content = {k: v for k, v in d.items() if k not in params_names}
        retval = {k: v for k, v in d.items() if k in params_names}
        retval["content"] = content
        retval["type"] = self.type
        return retval


# message type that asks user something


@dataclass
class AskingMessage(IOMessage): ...


# type of output messages
@dataclass
class TextMessage(IOMessage):
    body: Optional[str] = None

    def __post_init__(self) -> None:
        """Set the default value for the `type` attribute."""
        if self.type is None:
            self.type = "text_message"


@dataclass
class SuggestedFunctionCall(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    arguments: dict[str, Any] = field(default_factory=dict)


@dataclass
class FunctionCallExecution(IOMessage):
    function_name: Optional[str] = None
    call_id: Optional[str] = None
    retval: Any = None


# types of input messages
@dataclass
class TextInput(AskingMessage):
    prompt: Optional[str] = None
    suggestions: list[str] = field(default_factory=list)
    password: bool = False


@dataclass
class MultipleChoice(AskingMessage):
    prompt: Optional[str] = None
    choices: list[str] = field(default_factory=list)
    default: Optional[str] = None
    single: bool = True
    # todo: add validation


@dataclass
class SystemMessage(IOMessage):
    message: dict[str, Any] = field(default_factory=dict)


@dataclass
class WorkflowStarted(IOMessage):
    name: Optional[str] = None
    description: Optional[str] = None
    params: dict[str, Any] = field(default_factory=dict)


@dataclass
class WorkflowCompleted(IOMessage):
    result: Optional[str] = None


@dataclass
class Error(IOMessage):
    short: Optional[str] = None
    long: Optional[str] = None


@dataclass
class KeepAlive(IOMessage): ...


class MessageProcessorProtocol(Protocol):
    def process_message(self, message: IOMessage) -> Optional[str]: ...

    def text_message(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_message specific parameters
        body: Optional[str] = None,
    ) -> Optional[str]: ...

    def suggested_function_call(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # suggested_function_call specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        arguments: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def function_call_execution(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # function_call_execution specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        retval: Any = None,
    ) -> Optional[str]: ...

    def text_input(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_input specific parameters
        prompt: Optional[str] = None,
        suggestions: Optional[list[str]] = None,
        password: bool = False,
    ) -> Optional[str]: ...

    def multiple_choice(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # multiple_choice specific parameters
        prompt: Optional[str] = None,
        choices: Optional[list[str]] = None,
        default: Optional[str] = None,
        single: bool = True,
    ) -> Optional[str]: ...

    def system_message(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # system_message specific parameters
        message: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def workflow_started(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_started specific parameters
        name: Optional[str] = None,
        description: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Optional[str]: ...

    def workflow_completed(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_completed specific parameters
        result: Optional[str] = None,
    ) -> Optional[str]: ...

    def error(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # error specific parameters
        short: Optional[str] = None,
        long: Optional[str] = None,
    ) -> Optional[str]: ...

    def keep_alive(
        self,
        # common parameters for all messages
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
    ) -> Optional[str]: ...


class MessageProcessorMixin(ABC):
    def visit(self, message: IOMessage) -> Optional[str]:
        method_name = f"visit_{message.type}"
        method = getattr(self, method_name, self.visit_default)
        return method(message)

    @abstractmethod
    def visit_default(self, message: IOMessage) -> Optional[str]: ...

    def visit_text_message(self, message: TextMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_suggested_function_call(
        self, message: SuggestedFunctionCall
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_function_call_execution(
        self, message: FunctionCallExecution
    ) -> Optional[str]:
        return self.visit_default(message)

    def visit_text_input(self, message: TextInput) -> Optional[str]:
        return self.visit_default(message)

    def visit_multiple_choice(self, message: MultipleChoice) -> Optional[str]:
        return self.visit_default(message)

    def visit_system_message(self, message: SystemMessage) -> Optional[str]:
        return self.visit_default(message)

    def visit_keep_alive(self, message: KeepAlive) -> Optional[str]:
        return self.visit_default(message)

    def visit_workflow_started(self, message: WorkflowStarted) -> Optional[str]:
        return self.visit_default(message)

    def visit_workflow_completed(self, message: WorkflowCompleted) -> Optional[str]:
        return self.visit_default(message)

    def visit_error(self, message: Error) -> Optional[str]:
        return self.visit_default(message)

    def process_message(self, message: IOMessage) -> Optional[str]:
        try:
            return self.visit(message)
        except Exception as e:
            # log the error and return None
            logger.error(f"Error processing message ({message}): {e}", exc_info=True)
            return None

    def text_message(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        body: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            TextMessage(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                body=body,
            )
        )

    def suggested_function_call(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        arguments: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        arguments = arguments or {}
        return self.process_message(
            SuggestedFunctionCall(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                function_name=function_name,
                call_id=call_id,
                arguments=arguments,
            )
        )

    def function_call_execution(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        retval: Any = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            FunctionCallExecution(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                function_name=function_name,
                call_id=call_id,
                retval=retval,
            )
        )

    def text_input(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        prompt: Optional[str] = None,
        suggestions: Optional[list[str]] = None,
        password: bool = False,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        suggestions = suggestions or []
        return self.process_message(
            TextInput(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                prompt=prompt,
                suggestions=suggestions,
                password=password,
            )
        )

    def multiple_choice(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        prompt: Optional[str] = None,
        choices: Optional[list[str]] = None,
        default: Optional[str] = None,
        single: bool = True,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        choices = choices or []
        return self.process_message(
            MultipleChoice(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                prompt=prompt,
                choices=choices,
                default=default,
                single=single,
            )
        )

    def system_message(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        message: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        message = message or {}
        return self.process_message(
            SystemMessage(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                message=message,
            )
        )

    def workflow_started(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        params = params or {}
        return self.process_message(
            WorkflowStarted(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                name=name,
                description=description,
                params=params,
            )
        )

    def workflow_completed(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        result: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            WorkflowCompleted(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                result=result,
            )
        )

    def error(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        short: Optional[str] = None,
        long: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            Error(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
                short=short,
                long=long,
            )
        )

    def keep_alive(
        self,
        workflow_uuid: str,
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
    ) -> Optional[str]:
        uuid = uuid or str(uuid4().hex)
        return self.process_message(
            KeepAlive(
                sender=sender,
                recipient=recipient,
                auto_reply=auto_reply,
                uuid=uuid,
                workflow_uuid=workflow_uuid,
            )
        )


class InputResponseModel(BaseModel):
    msg: str
    question_uuid: Optional[UUID] = None
    error: bool = False


class InitiateWorkflowModel(BaseModel):
    user_id: Optional[str] = None
    workflow_uuid: UUID
    name: str
    params: dict[str, Any]



from collections.abc import Iterable

import mesop as me

from ...base import ProviderProtocol
from ...logging import get_logger
from .data_model import State
from .mesop import MesopMessage, MesopUI, run_workflow_mesop

logger = get_logger(__name__)


def send_prompt_to_autogen(
    provider: ProviderProtocol, name: str
) -> Iterable[MesopMessage]:
    ui = run_workflow_mesop(provider, name=name)
    if not isinstance(ui.ui_base, MesopUI):  # pragma: no cover
        logger.error("")
        raise RuntimeError(f"Expected MesopUI, got {type(ui.ui_base)}")
    mesop_ui: MesopUI = ui.ui_base

    state = me.state(State)
    state.conversation.fastagency = ui.ui_base.id
    return mesop_ui.get_message_stream()


def send_user_feedback_to_autogen(user_response: str) -> Iterable[MesopMessage]:
    state = me.state(State)
    mesop_id = state.conversation.fastagency
    mesop_io = MesopUI.get_conversation(mesop_id)
    mesop_io.respond(user_response)
    return mesop_io.get_message_stream()


def get_more_messages() -> Iterable[MesopMessage]:
    state = me.state(State)
    mesop_id = state.conversation.fastagency
    mesop_io = MesopUI.get_conversation(mesop_id)
    return mesop_io.get_message_stream()



from ...helpers import check_imports

check_imports(["fastapi_code_generator", "fastapi", "requests"], "openapi")

from .patch_fastapi_code_generator import (  # noqa: E402
    patch_function_name_parsing,
    patch_generate_code,
    patch_parse_schema,
)

patch_parse_schema()
patch_function_name_parsing()
patch_generate_code()

from .client import OpenAPI  # noqa: E402

__all__ = ["OpenAPI"]



import logging

__all__ = ["get_logger"]


def get_logger(
    name: str, *, level: int = logging.INFO, use_timestamp: bool = True
) -> logging.Logger:
    """Return a logger with the specified name.

    Args:
        name (str): The name of the logger
        level (int): The level of the logger
        use_timestamp (bool): Whether to use a timestamp in the log messages
    """
    # Get the logger
    logger = logging.getLogger(__name__)
    logger.setLevel(level)

    # Check if the logger already has handlers
    if not logger.handlers:
        # Create a stream handler
        handler = logging.StreamHandler()
        handler.setLevel(level)

        # Create a formatter and set it for the handler
        formatter = logging.Formatter(
            f"{'%(asctime)s' if use_timestamp else ''} [%(levelname)s] %(message)s"
        )
        handler.setFormatter(formatter)

        # Add the handler to the logger
        logger.addHandler(handler)

    return logger



import importlib
import inspect
import json
import re
import shutil
import sys
from collections.abc import Iterable, Iterator, Mapping
from contextlib import contextmanager
from functools import wraps
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, Any, Callable, Literal, Optional, Union

import fastapi
import requests
from datamodel_code_generator import DataModelType
from fastapi_code_generator.__main__ import generate_code
from pydantic_core import PydanticUndefined

from fastagency.helpers import optional_temp_path

from .fastapi_code_generator_helpers import patch_get_parameter_type
from .security import BaseSecurity, BaseSecurityParameters

if TYPE_CHECKING:
    from autogen.agentchat import ConversableAgent

__all__ = ["OpenAPI"]


@contextmanager
def add_to_globals(new_globals: dict[str, Any]) -> Iterator[None]:
    old_globals: dict[str, Any] = {}
    try:
        for key, value in new_globals.items():
            if key in globals():
                old_globals[key] = globals()[key]
            globals()[key] = value
        yield
    finally:
        for key, value in old_globals.items():
            globals()[key] = value


class OpenAPI:
    def __init__(
        self, servers: list[dict[str, Any]], title: Optional[str] = None, **kwargs: Any
    ) -> None:
        """Proxy class to generate client from OpenAPI schema."""
        self.servers = servers
        self.title = title
        self.kwargs = kwargs
        self.registered_funcs: list[Callable[..., Any]] = []
        self.globals: dict[str, Any] = {}

        self.security: dict[str, list[BaseSecurity]] = {}
        self.security_params: dict[Optional[str], BaseSecurityParameters] = {}

    @staticmethod
    def _convert_camel_case_within_braces_to_snake(text: str) -> str:
        # Function to convert camel case to snake case
        def camel_to_snake(match: re.Match[str]) -> str:
            return re.sub(r"(?<!^)(?=[A-Z])", "_", match.group(1)).lower()

        # Find all occurrences inside curly braces and apply camel_to_snake
        result = re.sub(
            r"\{([a-zA-Z0-9]+)\}", lambda m: "{" + camel_to_snake(m) + "}", text
        )

        return result

    @staticmethod
    def _get_params(
        path: str, func: Callable[..., Any]
    ) -> tuple[set[str], set[str], Optional[str], bool]:
        sig = inspect.signature(func)

        params_names = set(sig.parameters.keys())

        path_params = set(re.findall(r"\{(.+?)\}", path))
        if not path_params.issubset(params_names):
            raise ValueError(f"Path params {path_params} not in {params_names}")

        body = "body" if "body" in params_names else None

        security = "security" in params_names

        q_params = set(params_names) - path_params - {body} - {"security"}

        return q_params, path_params, body, security

    def _process_params(
        self, path: str, func: Callable[[Any], Any], **kwargs: Any
    ) -> tuple[str, dict[str, Any], dict[str, Any]]:
        path = OpenAPI._convert_camel_case_within_braces_to_snake(path)
        q_params, path_params, body, security = OpenAPI._get_params(path, func)

        expanded_path = path.format(**{p: kwargs[p] for p in path_params})

        url = self.servers[0]["url"] + expanded_path

        body_dict = {}
        if body and body in kwargs:
            body_value = kwargs[body]
            if isinstance(body_value, dict):
                body_dict = {"json": body_value}
            elif hasattr(body_value, "model_dump"):
                body_dict = {"json": body_value.model_dump()}
            else:
                body_dict = {"json": body_value.dict()}

        body_dict["headers"] = {"Content-Type": "application/json"}
        if security:
            q_params, body_dict = kwargs["security"].add_security(q_params, body_dict)
            # body_dict["headers"][security] = kwargs["security"]

        params = {k: v for k, v in kwargs.items() if k in q_params}

        return url, params, body_dict

    def set_security_params(
        self, security_params: BaseSecurityParameters, name: Optional[str] = None
    ) -> None:
        if name is not None:
            security = self.security.get(name)
            if security is None:
                raise ValueError(f"Security is not set for '{name}'")

            for match_security in security:
                if match_security.accept(security_params):
                    break
            else:
                raise ValueError(
                    f"Security parameters {security_params} do not match security {security}"
                )

        self.security_params[name] = security_params

    def _get_matching_security(
        self, security: list[BaseSecurity], security_params: BaseSecurityParameters
    ) -> BaseSecurity:
        # check if security matches security parameters
        for match_security in security:
            if match_security.accept(security_params):
                return match_security
        raise ValueError(
            f"Security parameters {security_params} does not match any given security {security}"
        )

    def _get_security_params(
        self, name: str
    ) -> tuple[Optional[BaseSecurityParameters], Optional[BaseSecurity]]:
        # check if security is set for the method
        security = self.security.get(name)
        if not security:
            return None, None

        security_params = self.security_params.get(name)
        if security_params is None:
            # check if default security parameters are set
            security_params = self.security_params.get(None)
            if security_params is None:
                raise ValueError(
                    f"Security parameters are not set for {name} and there are no default security parameters"
                )

        match_security = self._get_matching_security(security, security_params)

        return security_params, match_security

    def _request(
        self,
        method: Literal["put", "get", "post", "head", "delete"],
        path: str,
        description: Optional[str] = None,
        security: Optional[list[BaseSecurity]] = None,
        **kwargs: Any,
    ) -> Callable[..., dict[str, Any]]:
        def decorator(func: Callable[..., Any]) -> Callable[..., dict[str, Any]]:
            name = func.__name__

            if security is not None:
                self.security[name] = security

            @wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> dict[str, Any]:
                url, params, body_dict = self._process_params(path, func, **kwargs)

                security = self.security.get(name)
                if security is not None:
                    security_params, matched_security = self._get_security_params(name)
                    if security_params is None:
                        raise ValueError(
                            f"Security parameters are not set for '{name}'"
                        )
                    else:
                        security_params.apply(params, body_dict, matched_security)  # type: ignore [arg-type]

                response = getattr(requests, method)(url, params=params, **body_dict)
                return response.json()  # type: ignore [no-any-return]

            wrapper._description = (  # type: ignore [attr-defined]
                description or func.__doc__.strip()
                if func.__doc__ is not None
                else None
            )

            self.registered_funcs.append(wrapper)

            return wrapper

        return decorator  # type: ignore [return-value]

    def put(self, path: str, **kwargs: Any) -> Callable[..., dict[str, Any]]:
        return self._request("put", path, **kwargs)

    def get(self, path: str, **kwargs: Any) -> Callable[..., dict[str, Any]]:
        return self._request("get", path, **kwargs)

    def post(self, path: str, **kwargs: Any) -> Callable[..., dict[str, Any]]:
        return self._request("post", path, **kwargs)

    def delete(self, path: str, **kwargs: Any) -> Callable[..., dict[str, Any]]:
        return self._request("delete", path, **kwargs)

    def head(self, path: str, **kwargs: Any) -> Callable[..., dict[str, Any]]:
        return self._request("head", path, **kwargs)

    @classmethod
    def _get_template_dir(cls) -> Path:
        path = Path(__file__).parents[3] / "templates"
        if not path.exists():
            raise RuntimeError(f"Template directory {path.resolve()} not found.")
        return path

    @classmethod
    def generate_code(
        cls,
        input_text: str,
        output_dir: Path,
        disable_timestamp: bool = False,
        custom_visitors: Optional[list[Path]] = None,
    ) -> str:
        if custom_visitors is None:
            custom_visitors = []
        custom_visitors.append(Path(__file__).parent / "security_schema_visitor.py")

        with patch_get_parameter_type():
            generate_code(
                input_name="openapi.json",
                input_text=input_text,
                encoding="utf-8",
                output_dir=output_dir,
                template_dir=cls._get_template_dir(),
                disable_timestamp=disable_timestamp,
                custom_visitors=custom_visitors,
                output_model_type=DataModelType.PydanticV2BaseModel,
            )
            # Use unique file name for main.py
            main_name = f"main_{output_dir.name}"
            main_path = output_dir / f"{main_name}.py"
            shutil.move(output_dir / "main.py", main_path)

            # Change "from models import" to "from models_unique_name import"
            with main_path.open("r") as f:
                main_py_code = f.read()
            main_py_code = main_py_code.replace(
                "from .models import", f"from models_{output_dir.name} import"
            )

            with main_path.open("w") as f:
                f.write(main_py_code)

            # Use unique file name for models.py
            models_name = f"models_{output_dir.name}"
            models_path = output_dir / f"{models_name}.py"
            shutil.move(output_dir / "models.py", models_path)

            return main_name

    def set_globals(self, main: ModuleType, suffix: str) -> None:
        xs = {k: v for k, v in main.__dict__.items() if not k.startswith("__")}
        self.globals = {
            k: v
            for k, v in xs.items()
            if hasattr(v, "__module__")
            and v.__module__ in [f"models_{suffix}", "typing"]
        }

    @classmethod
    def create(
        cls,
        openapi_json: Optional[str] = None,
        openapi_url: Optional[str] = None,
        client_source_path: Optional[str] = None,
        servers: Optional[list[dict[str, Any]]] = None,
    ) -> "OpenAPI":
        if (openapi_json is None) == (openapi_url is None):
            raise ValueError("Either openapi_json or openapi_url should be provided")

        if openapi_json is None and openapi_url is not None:
            with requests.get(openapi_url, timeout=10) as response:
                response.raise_for_status()
                openapi_json = response.text

        if servers:
            openapi_parsed = json.loads(openapi_json)  # type: ignore [arg-type]
            openapi_parsed["servers"] = servers
            openapi_json = json.dumps(openapi_parsed)

        with optional_temp_path(client_source_path) as td:
            suffix = td.name

            main_name = cls.generate_code(
                input_text=openapi_json,  # type: ignore [arg-type]
                output_dir=td,
            )
            # add td to sys.path
            try:
                sys.path.append(str(td))
                main = importlib.import_module(main_name, package=td.name)  # nosemgrep
            finally:
                sys.path.remove(str(td))

            client: OpenAPI = main.app  # type: ignore [attr-defined]
            client.set_globals(main, suffix=suffix)

            return client

    def _get_functions_to_register(
        self,
        functions: Optional[
            Iterable[Union[str, Mapping[str, Mapping[str, str]]]]
        ] = None,
    ) -> dict[Callable[..., Any], dict[str, Union[str, None]]]:
        if functions is None:
            return {
                f: {"name": None, "description": None} for f in self.registered_funcs
            }

        functions_with_name_desc: dict[str, dict[str, Union[str, None]]] = {}

        for f in functions:
            if isinstance(f, str):
                functions_with_name_desc[f] = {"name": None, "description": None}
            elif isinstance(f, dict):
                functions_with_name_desc.update(
                    {
                        k: {
                            "name": v.get("name", None),
                            "description": v.get("description", None),
                        }
                        for k, v in f.items()
                    }
                )
            else:
                raise ValueError(f"Invalid type {type(f)} for function {f}")

        funcs_to_register: dict[Callable[..., Any], dict[str, Union[str, None]]] = {
            f: functions_with_name_desc[f.__name__]
            for f in self.registered_funcs
            if f.__name__ in functions_with_name_desc
        }
        missing_functions = set(functions_with_name_desc.keys()) - {
            f.__name__ for f in funcs_to_register
        }
        if missing_functions:
            raise ValueError(
                f"Following functions {missing_functions} are not valid functions"
            )

        return funcs_to_register

    @staticmethod
    def _remove_pydantic_undefined_from_tools(
        tools: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        for tool in tools:
            if "function" not in tool:
                continue

            function = tool["function"]
            if (
                "parameters" not in function
                or "properties" not in function["parameters"]
            ):
                continue

            required = function["parameters"].get("required", [])
            for param_name, param_value in function["parameters"]["properties"].items():
                if "default" not in param_value:
                    continue

                default = param_value.get("default")
                if (
                    isinstance(default, (fastapi.params.Path, fastapi.params.Query))
                    and param_value["default"].default is PydanticUndefined
                ):
                    param_value.pop("default")
                    # We removed the default value, so we need to add the parameter to the required list
                    if param_name not in required:
                        required.append(param_name)

        return tools

    def _register_for_llm(
        self,
        agent: "ConversableAgent",
        functions: Optional[
            Iterable[Union[str, Mapping[str, Mapping[str, str]]]]
        ] = None,
    ) -> None:
        funcs_to_register = self._get_functions_to_register(functions)

        with add_to_globals(self.globals):
            for f, v in funcs_to_register.items():
                agent.register_for_llm(name=v["name"], description=v["description"])(f)

            agent.llm_config["tools"] = OpenAPI._remove_pydantic_undefined_from_tools(
                agent.llm_config["tools"]
            )

    def _register_for_execution(
        self,
        agent: "ConversableAgent",
        functions: Optional[
            Iterable[Union[str, Mapping[str, Mapping[str, str]]]]
        ] = None,
    ) -> None:
        funcs_to_register = self._get_functions_to_register(functions)

        for f, v in funcs_to_register.items():
            agent.register_for_execution(name=v["name"])(f)

    def get_functions(self) -> list[str]:
        return [f.__name__ for f in self.registered_funcs]



from collections.abc import Iterator
from typing import Callable, Optional

import mesop as me

from ....logging import get_logger
from ..data_model import State
from ..styles import MesopTextInputInnerStyles

# Get the logger
logger = get_logger(__name__)


def input_text(  # noqa: C901
    on_input: Callable[[str], Iterator[None]],
    *,
    key: str,
    disabled: bool = False,
    value: Optional[str] = None,
    style: MesopTextInputInnerStyles,
) -> None:
    def on_click(e: me.ClickEvent) -> Iterator[None]:
        state = me.state(State)
        output_key = get_output_key()
        inp = getattr(state, output_key)
        clear_in_out()
        yield from on_input(inp)

    def on_newline(e: me.TextareaShortcutEvent) -> Iterator[None]:
        state = me.state(State)
        input_key = get_input_key()
        setattr(state, input_key, e.value + "\n")
        yield

    def on_submit(e: me.TextareaShortcutEvent) -> Iterator[None]:
        clear_in_out()
        yield from on_input(e.value)

    def on_blur(e: me.InputBlurEvent) -> None:
        if disabled or e.key != key_num:
            return
        state = me.state(State)
        input_key, output_key = get_in_out_keys()
        setattr(state, input_key, e.value)
        setattr(state, output_key, e.value)

    def get_input_key() -> str:
        return f"{key}_input"

    def get_output_key() -> str:
        return f"{key}_output"

    def get_in_out_keys() -> list[str]:
        return [get_input_key(), get_output_key()]

    def clear_in_out() -> None:
        input_key, output_key = get_in_out_keys()
        setattr(state, input_key, "")
        setattr(state, output_key, "")

    state = me.state(State)
    key_num = f"{key}{len(state.conversation.messages)}"
    with me.box(style=style.box):
        if disabled:
            in_value = value
            key_num = f"{key}disabled{len(state.conversation.messages)}"
        else:
            input_key = get_input_key()
            in_value = getattr(state, input_key)
            key_num = f"{key}{len(state.conversation.messages)}"

        with me.box(style=me.Style(flex_grow=1)):
            me.native_textarea(
                on_blur=on_blur,
                key=key_num,
                autosize=True,
                min_rows=3,
                max_rows=10,
                readonly=disabled,
                shortcuts={
                    me.Shortcut(key="enter", shift=True): on_newline,
                    me.Shortcut(key="enter"): on_submit,
                },
                style=style.native_textarea,
                value=in_value,
            )

        with me.content_button(
            type="icon",
            on_click=on_click,
            disabled=disabled,
        ):
            me.icon("send")



import os
from typing import Any

from autogen.agentchat import ConversableAgent

from fastagency import UI, FastAgency
from fastagency.runtimes.autogen import AutoGenWorkflows
from fastagency.ui.mesop import MesopUI

llm_config = {
    "config_list": [
        {
            "model": "gpt-4o-mini",
            "api_key": os.getenv("OPENAI_API_KEY"),
        }
    ],
    "temperature": 0.8,
}


wf = AutoGenWorkflows()


@wf.register(name="simple_learning", description="Student and teacher learning chat")
def simple_workflow(ui: UI, params: dict[str, Any]) -> str:
    initial_message = ui.text_input(
        sender="Workflow",
        recipient="User",
        prompt="I can help you learn about mathematics. What subject you would like to explore?",
    )
    student_agent = ConversableAgent(
        name="Student_Agent",
        system_message="You are a student willing to learn.",
        llm_config=llm_config,
    )
    teacher_agent = ConversableAgent(
        name="Teacher_Agent",
        system_message="You are a math teacher.",
        llm_config=llm_config,
    )

    chat_result = student_agent.initiate_chat(
        teacher_agent,
        message=initial_message,
        summary_method="reflection_with_llm",
        max_turns=5,
    )

    return chat_result.summary


app = FastAgency(provider=wf, ui=MesopUI())

# start with the following command
# gunicorn main_mesop:app -b 0.0.0.0:8888 --reload



from typing import Any, Optional, Union

from autogen import AssistantAgent, ConversableAgent

from ..tools import WebSurferTool


class WebSurferAgent(AssistantAgent):  # type: ignore[misc]
    def __init__(
        self,
        *args: Any,
        name: str,
        llm_config: dict[str, Any],
        summarizer_llm_config: dict[str, Any],
        executor: Union[ConversableAgent, list[ConversableAgent]],
        system_message: str = "You are a web surfer",
        bing_api_key: Optional[str] = None,
        **kwargs: Any,
    ):
        """Initialize the WebSurferAgent.

        Args:
            *args (Any): The positional arguments.
            name (str): The name of the agent.
            llm_config (dict[str, Any]): The LLM configuration.
            summarizer_llm_config (dict[str, Any]): The summarizer LLM configuration.
            executor (Union[ConversableAgent, list[ConversableAgent]]): The executor agent(s).
            system_message (str): The system message.
            bing_api_key (Optional[str]): The Bing API key
            **kwargs (Any): The keyword arguments.
        """
        super().__init__(
            *args,
            name=name,
            system_message=system_message,
            llm_config=llm_config,
            **kwargs,
        )
        self.web_surfer_tool = WebSurferTool(
            name_prefix="Web_Surfer",
            llm_config=llm_config,
            summarizer_llm_config=summarizer_llm_config,
            bing_api_key=bing_api_key,
        )
        self.web_surfer_tool.register(caller=self, executor=executor)



import inspect
from collections.abc import Awaitable, Generator, Iterable, Iterator, Mapping
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Protocol,
    TypeVar,
    Union,
    runtime_checkable,
)

from .logging import get_logger
from .messages import IOMessage, MessageProcessorProtocol

if TYPE_CHECKING:
    from fastagency.api.openapi import OpenAPI

__all__ = [
    "UIBase",
    "WSGIProtocol",
    "ASGIProtocol",
    "ProviderProtocol",
    "WorkflowsProtocol",
    "AdapterProtocol",
    "Runnable",
    "Workflow",
    "Agent",
]

logger = get_logger(__name__)


@runtime_checkable
class UIBase(MessageProcessorProtocol, Protocol):
    @contextmanager
    def create(self, app: "Runnable", import_string: str) -> Iterator[None]: ...

    def start(
        self,
        *,
        app: "Runnable",
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None: ...

    def create_workflow_ui(self, workflow_uuid: str) -> "UI": ...

    # def process_streaming_message(
    #     self, message: IOStreamingMessage
    # ) -> Optional[str]: ...


class CreateWorkflowUIMixin:
    def create_workflow_ui(self: UIBase, workflow_uuid: str) -> "UI":
        return UI(uibase=self, workflow_uuid=workflow_uuid)


class UI:
    def __init__(self, uibase: UIBase, workflow_uuid: str) -> None:
        if workflow_uuid is None:
            logger.error("workflow_uuid must be provided")
            raise ValueError("workflow_uuid must be provided")
        self._ui_base = uibase
        self._workflow_uuid = workflow_uuid

    @property
    def workflow_uuid(self) -> str:
        return self._workflow_uuid

    @property
    def ui_base(self) -> UIBase:
        return self._ui_base

    def process_message(self, message: IOMessage) -> Optional[str]:
        return self._ui_base.process_message(message)

    def text_message(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_message specific parameters
        body: Optional[str] = None,
    ) -> Optional[str]:
        return self._ui_base.text_message(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            body=body,
        )

    def suggested_function_call(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # suggested_function_call specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        arguments: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        return self._ui_base.suggested_function_call(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            function_name=function_name,
            call_id=call_id,
            arguments=arguments,
        )

    def function_call_execution(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # function_call_execution specific parameters
        function_name: Optional[str] = None,
        call_id: Optional[str] = None,
        retval: Any = None,
    ) -> Optional[str]:
        return self._ui_base.function_call_execution(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            function_name=function_name,
            call_id=call_id,
            retval=retval,
        )

    def text_input(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # text_input specific parameters
        prompt: Optional[str] = None,
        suggestions: Optional[list[str]] = None,
        password: bool = False,
    ) -> Optional[str]:
        return self._ui_base.text_input(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            prompt=prompt,
            suggestions=suggestions,
            password=password,
        )

    def multiple_choice(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # multiple_choice specific parameters
        prompt: Optional[str] = None,
        choices: Optional[list[str]] = None,
        default: Optional[str] = None,
        single: bool = True,
    ) -> Optional[str]:
        return self._ui_base.multiple_choice(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            prompt=prompt,
            choices=choices,
            default=default,
            single=single,
        )

    def system_message(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # system_message specific parameters
        message: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        return self._ui_base.system_message(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            message=message,
        )

    def workflow_started(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_started specific parameters
        name: Optional[str] = None,
        description: Optional[str] = None,
        params: Optional[dict[str, Any]] = None,
    ) -> Optional[str]:
        return self._ui_base.workflow_started(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            name=name,
            description=description,
            params=params,
        )

    def workflow_completed(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # workflow_completed specific parameters
        result: Optional[str] = None,
    ) -> Optional[str]:
        return self._ui_base.workflow_completed(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            result=result,
        )

    def error(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
        # error specific parameters
        short: Optional[str] = None,
        long: Optional[str] = None,
    ) -> Optional[str]:
        return self._ui_base.error(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
            short=short,
            long=long,
        )

    def keep_alive(
        self,
        # common parameters for all messages
        sender: Optional[str] = None,
        recipient: Optional[str] = None,
        auto_reply: bool = False,
        uuid: Optional[str] = None,
    ) -> Optional[str]:
        return self._ui_base.keep_alive(
            workflow_uuid=self.workflow_uuid,
            sender=sender,
            recipient=recipient,
            auto_reply=auto_reply,
            uuid=uuid,
        )


@runtime_checkable
class WSGIProtocol(Protocol):
    def handle_wsgi(
        self,
        app: "Runnable",
        environ: dict[str, Any],
        start_response: Callable[..., Any],
    ) -> list[bytes]: ...


@runtime_checkable
class ASGIProtocol(Protocol):
    async def handle_asgi(
        self,
        app: "Runnable",
        scope: dict[str, Any],
        receive: Callable[[dict[str, Any]], Awaitable[None]],
        send: Callable[[dict[str, Any]], Awaitable[None]],
    ) -> None: ...


# signature of a function decorated with @wf.register
Workflow = TypeVar("Workflow", bound=Callable[[UI, dict[str, Any]], str])


Agent = TypeVar("Agent")


@runtime_checkable
class ProviderProtocol(Protocol):
    def run(
        self,
        name: str,
        ui: UI,
        user_id: Optional[str] = None,
        **kwargs: Any,
    ) -> str: ...

    """Run a workflow.

    Creates a new workflow and assigns it workflow_uuid. Then it calls the
    workflow function (function decorated with @wf.register) with the given
    ui and workflow_uuid.

    Args:
        name (str): The name of the workflow to run.
        ui (UI): The UI object to use.
        **kwargs: Additional parameters to pass to the workflow function.
    """

    @property
    def names(self) -> list[str]: ...

    def get_description(self, name: str) -> str: ...


@runtime_checkable
class WorkflowsProtocol(ProviderProtocol, Protocol):
    def register(
        self, name: str, description: str
    ) -> Callable[[Workflow], Workflow]: ...

    def register_api(
        self,
        api: "OpenAPI",
        callers: Union[Agent, Iterable[Agent]],
        executors: Union[Agent, Iterable[Agent]],
        functions: Optional[
            Union[str, Iterable[Union[str, Mapping[str, Mapping[str, str]]]]]
        ] = None,
    ) -> None: ...


def check_register_decorator(func: Workflow) -> None:
    # get names of all parameters in the function signature
    sig = inspect.signature(func)
    params = list(sig.parameters.keys())
    if params != ["ui", "params"]:
        raise ValueError(
            f"Expected function signature to be 'def func(ui: UI, workflow_uuid: str, params: dict[str, Any]) -> str', got {sig}"
        )


@runtime_checkable
class AdapterProtocol(Protocol):
    @classmethod
    def create_provider(*args: Any, **kwargs: Any) -> ProviderProtocol: ...


@runtime_checkable
class Runnable(Protocol):
    @contextmanager
    def create(self, import_string: str) -> Generator[None, None, None]: ...

    def start(
        self,
        *,
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None: ...

    @property
    def provider(self) -> ProviderProtocol: ...

    @property
    def ui(self) -> UIBase: ...

    @property
    def title(self) -> str: ...

    @property
    def description(self) -> str: ...



import os
from typing import Annotated, Any, Optional

from autogen import register_function
from autogen.agentchat import ConversableAgent

from fastagency import UI, FastAgency
from fastagency.runtimes.autogen import AutoGenWorkflows
from fastagency.ui.mesop import MesopUI

wf = AutoGenWorkflows()

@wf.register(name="text_message", description="Text message test")
def many_multiple_choice_workflow(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.text_message(
        sender="Workflow",
        recipient="User",
        body="text message body"
    )
    return "Completed OK"  # type: ignore[no-any-return]

@wf.register(name="text_input_message", description="Text input message test")
def multiple_choice_single_workflow(
    ui: UI, params: dict[str, Any]
) -> str:
    answer = ui.text_input(
        sender="Workflow",
        recipient="User",
        prompt="Have you run any tests recently?"
    )
    return answer  # type: ignore[no-any-return]


@wf.register(name="multiple_choice_message_single", description="Multiple choice - single")
def multiple_choice_single_workflow(
    ui: UI, params: dict[str, Any]
) -> str:
    answer = ui.multiple_choice(
        sender="Workflow",
        recipient="User",
        prompt="When was the last time you have run automated tests?",
        choices=["I am currently running them", "yesterday", "Tests? What tests??"]
    )
    return f"you have chosen: {answer}"  # type: ignore[no-any-return]

@wf.register(name="multiple_choice_message_many", description="Multiple choice - many")
def multiple_choice_many_workflow(
    ui: UI, params: dict[str, Any]
) -> str:
    answer = ui.multiple_choice(
        sender="Workflow",
        recipient="User",
        prompt="When was the last time you have run automated tests?",
        choices=["I am currently running them", "yesterday", "Tests? What tests??"],
        single=False
    )
    return f"you have chosen: {answer}"  # type: ignore[no-any-return]


@wf.register(name="suggested_function_call_message", description="Suggested Function Call Message")
def suggested_function_call_workflow(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.suggested_function_call(
        sender="Workflow",
        recipient="User",
        function_name="the name of the function"
    )
    return "Completed OK"  # type: ignore[no-any-return]

@wf.register(name="function_call_execution_message", description="Function Call Execution Message")
def function_call_execution(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.function_call_execution(
        sender="Workflow",
        recipient="User",
        function_name="the name of the function"
    )
    return "Completed OK"  # type: ignore[no-any-return]

@wf.register(name="error_message", description="Error Message")
def error_message(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.error(
        sender="Workflow",
        recipient="User",
        short="This is an Error in short form",
        long="This is an Error in somewhat longer form"
    )
    return "Completed OK"  # type: ignore[no-any-return]

@wf.register(name="workflow_started", description="Workflow started")
def workflow_started(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.workflow_started(
        sender="Workflow",
        recipient="User",
        name="_workflow_started_",
        description="The beginnings are delicate times...",
    )
    return "Completed OK"  # type: ignore[no-any-return]

@wf.register(name="workflow_completed", description="Workflow completed")
def workflow_completed(
    ui: UI, params: dict[str, Any]
) -> str:
    ui.workflow_completed(
        sender="Workflow",
        recipient="User",
        result="This workflow has completed",
    )
    return "Completed OK"  # type: ignore[no-any-return]


app = FastAgency(provider=wf, ui=MesopUI(), title="Learning Chat")



from pathlib import Path

from fastapi_code_generator.parser import OpenAPIParser
from fastapi_code_generator.visitor import Visitor

from fastagency.api.openapi.security import BaseSecurity


def custom_visitor(parser: OpenAPIParser, model_path: Path) -> dict[str, object]:
    if "securitySchemes" not in parser.raw_obj["components"]:
        return {}
    security_schemes = parser.raw_obj["components"]["securitySchemes"]
    server_url = parser.raw_obj["servers"][0]["url"]

    security_classes = []
    security_parameters = {}
    for k, v in security_schemes.items():
        v["server_url"] = server_url
        security_class = BaseSecurity.get_security_class(
            type=v["type"], schema_parameters=v
        )

        security_classes.append(security_class.__name__)

        security_parameters[k] = security_class.get_security_parameters(
            schema_parameters=v
        )

    return {
        "security_schemes": security_schemes,
        "security_classes": security_classes,
        "security_parameters": security_parameters,
    }


visit: Visitor = custom_visitor



## IONats part

import asyncio
import json
from collections.abc import Iterator
from contextlib import AsyncExitStack, contextmanager
from typing import Any, Callable, Optional
from uuid import UUID, uuid4

import requests
import websockets
from asyncer import asyncify, syncify
from fastapi import (
    APIRouter,
    Depends,
    HTTPException,
    Request,
    Response,
    WebSocket,
)
from fastapi.dependencies.utils import get_dependant, solve_dependencies
from pydantic import BaseModel

from fastagency.logging import get_logger

from ...base import (
    UI,
    CreateWorkflowUIMixin,
    ProviderProtocol,
    Runnable,
    UIBase,
    WorkflowsProtocol,
)
from ...exceptions import (
    FastAgencyConnectionError,
    FastAgencyFastAPIConnectionError,
    FastAgencyKeyError,
)
from ...messages import (
    AskingMessage,
    IOMessage,
    InitiateWorkflowModel,
    InputResponseModel,
    MessageProcessorMixin,
)

logger = get_logger(__name__)


class InititateChatModel(BaseModel):
    workflow_name: str
    workflow_uuid: str
    user_id: Optional[str]
    params: dict[str, Any]


class WorkflowInfo(BaseModel):
    name: str
    description: str


class FastAPIAdapter(MessageProcessorMixin, CreateWorkflowUIMixin):
    def __init__(
        self,
        provider: ProviderProtocol,
        *,
        initiate_workflow_path: str = "/fastagency/initiate_workflow",
        discovery_path: str = "/fastagency/discovery",
        ws_path: str = "/fastagency/ws",
        get_user_id: Optional[Callable[..., Optional[str]]] = None,
    ) -> None:
        """Provider for FastAPI.

        Args:
            provider (ProviderProtocol): The provider.
            initiate_workflow_path (str, optional): The initiate workflow path. Defaults to "/fastagency/initiate_workflow".
            discovery_path (str, optional): The discovery path. Defaults to "/fastagency/discovery".
            ws_path (str, optional): The websocket path. Defaults to "/fastagency/ws".
            get_user_id (Optional[Callable[[], Optional[UUID]]], optional): The get user id. Defaults to None.
        """
        self.provider = provider

        self.initiate_workflow_path = initiate_workflow_path
        self.discovery_path = discovery_path
        self.ws_path = ws_path

        self.get_user_id = get_user_id or (lambda: None)

        self.websockets: dict[str, WebSocket] = {}

        self.router = self.setup_routes()

    async def get_user_id_websocket(self, websocket: WebSocket) -> Optional[str]:
        def get_user_id_depends_stub(
            user_id: Optional[str] = Depends(self.get_user_id),
        ) -> Optional[str]:
            raise RuntimeError(
                "Stub get_user_id_depends_stub called"
            )  # pragma: no cover

        dependant = get_dependant(path="", call=get_user_id_depends_stub)

        try:
            async with AsyncExitStack() as cm:
                scope = websocket.scope
                scope["type"] = "http"

                solved_dependency = await solve_dependencies(
                    dependant=dependant,
                    request=Request(scope=scope),  # Inject the request here
                    body=None,
                    dependency_overrides_provider=None,
                    async_exit_stack=cm,
                    embed_body_fields=False,
                )
        except HTTPException as e:
            raise e
        finally:
            scope["type"] = "websocket"

        return solved_dependency.values["user_id"]  # type: ignore[no-any-return]

    def setup_routes(self) -> APIRouter:
        router = APIRouter()

        @router.post(self.initiate_workflow_path)
        async def initiate_chat(
            initiate_chat: InititateChatModel,
            user_id: Optional[str] = Depends(self.get_user_id),
        ) -> InitiateWorkflowModel:
            workflow_uuid: UUID = uuid4()

            init_msg = InitiateWorkflowModel(
                user_id=user_id,
                workflow_uuid=workflow_uuid,
                params=initiate_chat.params,
                name=initiate_chat.workflow_name,
            )

            return init_msg

        @router.websocket(self.ws_path)
        async def websocket_endpoint(
            websocket: WebSocket,
        ) -> None:
            try:
                user_id = await self.get_user_id_websocket(websocket)
            except HTTPException as e:
                headers = getattr(e, "headers", None)
                await websocket.send_denial_response(
                    Response(status_code=e.status_code, headers=headers)
                )
                return

            logger.info("Websocket connected")
            await websocket.accept()
            logger.info("Websocket accepted")

            init_msg_json = await websocket.receive_text()
            logger.info(f"Received message: {init_msg_json}")

            init_msg = InitiateWorkflowModel.model_validate_json(init_msg_json)

            workflow_uuid = init_msg.workflow_uuid.hex
            self.websockets[workflow_uuid] = websocket

            try:
                await asyncify(self.provider.run)(
                    name=init_msg.name,
                    ui=self.create_workflow_ui(workflow_uuid),
                    user_id=user_id if user_id else "None",
                    **init_msg.params,
                )
            except Exception as e:
                logger.error(f"Error in websocket_endpoint: {e}", stack_info=True)
            finally:
                self.websockets.pop(workflow_uuid)

        @router.get(
            self.discovery_path,
            responses={
                404: {"detail": "Key Not Found"},
                504: {"detail": "Unable to connect to provider"},
            },
        )
        def discovery(
            user_id: Optional[str] = Depends(self.get_user_id),
        ) -> list[WorkflowInfo]:
            try:
                names = self.provider.names
            except FastAgencyConnectionError as e:
                raise HTTPException(status_code=504, detail=str(e)) from e

            try:
                descriptions = [self.provider.get_description(name) for name in names]
            except FastAgencyKeyError as e:
                raise HTTPException(status_code=404, detail=str(e)) from e

            return [
                WorkflowInfo(name=name, description=description)
                for name, description in zip(names, descriptions)
            ]

        return router

    def visit_default(self, message: IOMessage) -> Optional[str]:
        async def a_visit_default(
            self: FastAPIAdapter, message: IOMessage
        ) -> Optional[str]:
            workflow_uuid = message.workflow_uuid
            if workflow_uuid not in self.websockets:
                logger.error(
                    f"Workflow {workflow_uuid} not found in websockets: {self.websockets}"
                )
                raise RuntimeError(
                    f"Workflow {workflow_uuid} not found in websockets: {self.websockets}"
                )
            websocket = self.websockets[workflow_uuid]  # type: ignore[index]
            await websocket.send_text(json.dumps(message.model_dump()))

            if isinstance(message, AskingMessage):
                response = await websocket.receive_text()
                return response
            return None

        return syncify(a_visit_default)(self, message)

    def create_subconversation(self) -> UIBase:
        return self

    @contextmanager
    def create(self, app: Runnable, import_string: str) -> Iterator[None]:
        raise NotImplementedError("create")

    def start(
        self,
        *,
        app: "Runnable",
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None:
        raise NotImplementedError("start")

    @classmethod
    def create_provider(
        cls,
        fastapi_url: str,
    ) -> ProviderProtocol:
        return FastAPIProvider(
            fastapi_url=fastapi_url,
        )


class FastAPIProvider(ProviderProtocol):
    def __init__(
        self,
        fastapi_url: str,
        initiate_workflow_path: str = "/fastagency/initiate_workflow",
        discovery_path: str = "/fastagency/discovery",
        ws_path: str = "/fastagency/ws",
    ) -> None:
        """Initialize the fastapi workflows."""
        self._workflows: dict[
            str, tuple[Callable[[WorkflowsProtocol, UIBase, str, str], str], str]
        ] = {}

        self.fastapi_url = (
            fastapi_url[:-1] if fastapi_url.endswith("/") else fastapi_url
        )
        self.ws_url = "ws" + self.fastapi_url[4:]

        self.is_broker_running: bool = False

        self.initiate_workflow_path = initiate_workflow_path
        self.discovery_path = discovery_path
        self.ws_path = ws_path

    def _send_initiate_chat_msg(
        self,
        workflow_name: str,
        workflow_uuid: str,
        user_id: Optional[str],
        params: dict[str, Any],
    ) -> InitiateWorkflowModel:
        msg = InititateChatModel(
            workflow_name=workflow_name,
            workflow_uuid=workflow_uuid,
            user_id=user_id,
            params=params,
        )

        payload = msg.model_dump()

        resp = requests.post(
            f"{self.fastapi_url}{self.initiate_workflow_path}", json=payload, timeout=5
        )
        logger.info(f"Initiate chat response: {resp.json()}")
        retval = InitiateWorkflowModel(**resp.json())
        return retval

    async def _publish_websocket_message(
        self,
        websocket: websockets.WebSocketClientProtocol,
        message: InputResponseModel,
    ) -> None:
        payload = message.model_dump_json()
        await websocket.send(payload)
        logger.info(f"Message sent to websocket ({websocket}): {message}")

    async def _run_websocket_subscriber(
        self,
        ui: UI,
        workflow_name: str,
        user_id: Optional[str],
        from_server_subject: str,
        to_server_subject: str,
        params: dict[str, Any],
    ) -> None:
        connect_url = f"{self.ws_url}{self.ws_path}"
        async with websockets.connect(connect_url) as websocket:
            init_workflow_msg = InitiateWorkflowModel(
                name=workflow_name,
                workflow_uuid=ui._workflow_uuid,
                user_id=user_id,
                params=params,
            )
            await websocket.send(init_workflow_msg.model_dump_json())

            while True:
                response = await websocket.recv()
                response = (
                    response.decode() if isinstance(response, bytes) else response
                )

                logger.info(f"Received message: {response}")

                msg = IOMessage.create(**json.loads(response))

                retval = await asyncify(ui.process_message)(msg)
                logger.info(f"Message {msg}: processed with response {retval}")

                if isinstance(msg, AskingMessage):
                    if retval is None:
                        logger.warning(
                            f"Message {msg}: response is None. Skipping response to websocket"
                        )
                    else:
                        await websocket.send(retval)
                        logger.info(
                            f"Message {msg}: response {retval} sent to websocket"
                        )

    def run(
        self,
        name: str,
        ui: UI,
        user_id: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        workflow_uuid = ui._workflow_uuid

        initiate_workflow = self._send_initiate_chat_msg(
            name, workflow_uuid=workflow_uuid, user_id=user_id, params=kwargs
        )
        user_id = initiate_workflow.user_id if initiate_workflow.user_id else "None"
        workflow_uuid = initiate_workflow.workflow_uuid.hex

        _from_server_subject = f"chat.client.messages.{user_id}.{workflow_uuid}"
        _to_server_subject = f"chat.server.messages.{user_id}.{workflow_uuid}"

        async def _setup_and_run() -> None:
            await self._run_websocket_subscriber(
                ui,
                name,
                user_id,
                _from_server_subject,
                _to_server_subject,
                kwargs,
            )

        async def run_lifespan() -> None:
            if not self.is_broker_running:
                self.is_broker_running = True
                await _setup_and_run()
            else:
                await _setup_and_run()

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        loop.run_until_complete(run_lifespan())

        return "FastAPIWorkflows.run() completed"

    def _get_workflow_info(self) -> list[dict[str, str]]:
        try:
            resp = requests.get(f"{self.fastapi_url}/{self.discovery_path}", timeout=15)
        except requests.exceptions.ConnectionError as e:
            raise FastAgencyFastAPIConnectionError(
                f"Unable to connect to FastAPI server at {self.fastapi_url}"
            ) from e
        if resp.status_code == 504:
            raise FastAgencyConnectionError(resp.json()["detail"])
        elif resp.status_code == 404:
            raise FastAgencyKeyError(resp.json()["detail"])
        return resp.json()  # type: ignore [no-any-return]

    def _get_names(self) -> list[str]:
        return [workflow["name"] for workflow in self._get_workflow_info()]

    def _get_description(self, name: str) -> str:
        return next(
            workflow["description"]
            for workflow in self._get_workflow_info()
            if workflow["name"] == name
        )

    @property
    def names(self) -> list[str]:
        return self._get_names()

    def get_description(self, name: str) -> str:
        return self._get_description(name)



import os
from typing import Any

from autogen.agentchat import ConversableAgent, UserProxyAgent

from fastagency import UI, FastAgency
from fastagency.api.openapi.client import OpenAPI
from fastagency.api.openapi.security import APIKeyHeader
from fastagency.runtimes.autogen import AutoGenWorkflows
from fastagency.ui.console import ConsoleUI

llm_config = {
    "config_list": [
        {
            "model": "gpt-4o-mini",
            "api_key": os.getenv("OPENAI_API_KEY"),
        }
    ],
    "temperature": 0.8,
}


wf = AutoGenWorkflows()


@wf.register(name="weatherman_workflow", description="Weatherman chat")
def simple_workflow(ui: UI, params: dict[str, Any]) -> str:
    user_proxy = UserProxyAgent(
        name="User_Proxy",
        human_input_mode="ALWAYS",
        llm_config=llm_config,
    )
    weatherman = ConversableAgent(
        name="Weatherman",
        system_message="You are a weatherman.",
        llm_config=llm_config,
    )

    weather_client = OpenAPI.create(
        openapi_url="https://weather.tools.fastagency.ai/openapi.json"
    )
    # Set global security params for all methods
    weather_client.set_security_params(
        APIKeyHeader.Parameters(value="secure weather key")
    )

    wf.register_api(
        api=weather_client,
        callers=user_proxy,
        executors=weatherman,
    )

    initial_message = ui.text_input(
        sender="Workflow",
        recipient="User",
        prompt="What would you like to find out about weather?",
    )

    chat_result = user_proxy.initiate_chat(
        weatherman,
        message=initial_message,
        summary_method="reflection_with_llm",
        max_turns=3,
    )

    return chat_result.summary


app = FastAgency(provider=wf, ui=ConsoleUI())



## IONats part

import asyncio
import os
import traceback
from collections.abc import AsyncIterator, Iterator
from contextlib import asynccontextmanager, contextmanager
from queue import Queue
from typing import TYPE_CHECKING, Any, Optional

from asyncer import asyncify, syncify
from faststream import FastStream, Logger
from faststream.nats import JStream, NatsBroker, NatsMessage
from nats.aio.client import Client as NatsClient
from nats.errors import NoServersError
from nats.js import JetStreamContext, api
from nats.js.errors import KeyNotFoundError, NoKeysError
from nats.js.kv import KeyValue

from ...base import UI, CreateWorkflowUIMixin, ProviderProtocol, Runnable, UIBase
from ...exceptions import FastAgencyNATSConnectionError, FastAgencyNATSKeyError
from ...logging import get_logger
from ...messages import (
    AskingMessage,
    IOMessage,
    InitiateWorkflowModel,
    InputResponseModel,
    MessageProcessorMixin,
    MultipleChoice,
    TextInput,
    TextMessage,
)

if TYPE_CHECKING:
    from faststream.nats.subscriber.asyncapi import AsyncAPISubscriber


logger = get_logger(__name__)

JETSTREAM = JStream(
    name="FastAgency",
    subjects=[
        # starts new conversation (can land on any worker)
        "chat.server.initiate_chat",
        # server requests input from client; chat.client.messages.<user_uuid>.<workflow_uuid>
        # we create this topic dynamically => client process consuming NATS can fix its worker
        "chat.client.messages.*.*",
        # server prints message to client; chat.server.messages.<user_uuid>.<workflow_uuid>
        # we create this topic dynamically and subscribe to it => worker is fixed
        "chat.server.messages.*.*",
        # discovery subject
        "discovery",
    ],
)


class NatsAdapter(MessageProcessorMixin, CreateWorkflowUIMixin):
    def __init__(
        self,
        provider: ProviderProtocol,
        *,
        nats_url: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
        super_conversation: Optional["NatsAdapter"] = None,
    ) -> None:
        """Provider for NATS.

        Args:
            provider (ProviderProtocol): The provider.
            nats_url (Optional[str], optional): The NATS URL. Defaults to None in which case 'nats://localhost:4222' is used.
            user (Optional[str], optional): The user. Defaults to None.
            password (Optional[str], optional): The password. Defaults to None.
            super_conversation (Optional["NatsProvider"], optional): The super conversation. Defaults to None.
        """
        self.provider = provider
        self.nats_url = nats_url or "nats://localhost:4222"
        self.user = user
        self.password = password
        self.queue: Queue = Queue()  # type: ignore[type-arg]

        self.broker = NatsBroker(self.nats_url, user=user, password=password)
        self.app = FastStream(self.broker)
        self.subscriber: "AsyncAPISubscriber"
        self._input_request_subject: str
        self._input_receive_subject: str

        self.super_conversation: Optional[NatsAdapter] = super_conversation
        self.sub_conversations: list[NatsAdapter] = []

        self._create_initiate_subscriber()

    async def _handle_input(
        self, body: InputResponseModel, msg: NatsMessage, logger: Logger
    ) -> None:
        """Handle input from the client by consuming messages from chat.server.messages.*.*.

        Args:
            body (InputResponseModel): The body of the message.
            msg (NatsMessage): The message object.
            logger (Logger): The logger object (gets injected)
        """
        logger.info(
            f"Received message in subject '{self._input_receive_subject}': {body}"
        )
        await msg.ack()
        self.queue.put(msg)

    async def _send_error_msg(self, e: Exception, logger: Logger) -> None:
        """Send an error message.

        Args:
            e (Exception): The exception.
            logger (Logger): The logger object (gets injected)
        """
        logger.error(f"Error in chat: {e}")
        logger.error(traceback.format_exc())

        error_msg = InputResponseModel(msg=str(e), error=True, question_uuid=None)
        await self.broker.publish(error_msg, self._input_request_subject)

    def _create_initiate_subscriber(self) -> None:
        @self.broker.subscriber(
            "chat.server.initiate_chat",
            stream=JETSTREAM,
            queue="initiate_workers",
            deliver_policy=api.DeliverPolicy("all"),
        )
        async def initiate_handler(
            body: InitiateWorkflowModel, msg: NatsMessage, logger: Logger
        ) -> None:
            """Initiate the handler.

            1. Subscribes to the chat.server.initiate_chat topic.
            2. When a message is consumed from the topic, it dynamically subscribes to the chat.server.messages.<user_uuid>.<workflow_uuid> topic.
            3. Starts the chat workflow after successfully subscribing to the chat.server.messages.<user_uuid>.<workflow_uuid> topic.

            Args:
                body (InitiateModel): The body of the message.
                msg (NatsMessage): The message object.
                logger (Logger): The logger object (gets injected)

            """
            await msg.ack()

            logger.info(
                f"Message in subject 'chat.server.initiate_chat': {body=} -> from process id {os.getpid()}"
            )
            user_id = body.user_id if body.user_id else "None"
            workflow_uuid = body.workflow_uuid.hex
            self._input_request_subject = (
                f"chat.client.messages.{user_id}.{workflow_uuid}"
            )
            self._input_receive_subject = (
                f"chat.server.messages.{user_id}.{workflow_uuid}"
            )

            # dynamically subscribe to the chat server
            subscriber = self.broker.subscriber(
                subject=self._input_receive_subject,
                stream=JETSTREAM,
                deliver_policy=api.DeliverPolicy("all"),
            )
            subscriber(self._handle_input)
            self.broker.setup_subscriber(subscriber)
            await subscriber.start()

            try:

                async def start_chat(
                    ui_base: UIBase,
                    provider: ProviderProtocol,
                    name: str,
                    params: dict[str, Any],
                    workflow_uuid: str,
                ) -> None:  # type: ignore [return]
                    def _start_chat(
                        ui_base: UIBase,
                        provider: ProviderProtocol,
                        name: str,
                        params: dict[str, Any],
                        workflow_uuid: str,
                    ) -> None:  # type: ignore [return]
                        ui: UI = ui_base.create_workflow_ui(workflow_uuid=workflow_uuid)
                        try:
                            provider.run(
                                name=name,
                                ui=ui,
                                **params,
                            )
                        except Exception as e:
                            logger.error(
                                f"Unexpecter error in NatsAdapter.start_chat: {e}",
                                stack_info=True,
                            )
                            ui.error(
                                sender="NatsAdapter",
                                short=f"Unexpected error: {e}",
                                long=traceback.format_exc(),
                            )
                            return

                    return await asyncify(_start_chat)(
                        ui_base, provider, name, params, workflow_uuid
                    )

                background_tasks = set()
                task = asyncio.create_task(
                    start_chat(
                        self, self.provider, body.name, body.params, workflow_uuid
                    )
                )  # type: ignore
                background_tasks.add(task)

                async def callback(t: asyncio.Task[Any]) -> None:
                    try:
                        background_tasks.discard(t)
                        await subscriber.close()
                    except Exception as e:
                        logger.error(f"Error in callback: {e}")
                        logger.error(traceback.format_exc())

                task.add_done_callback(lambda t: asyncio.create_task(callback(t)))

            except Exception as e:
                await self._send_error_msg(e, logger)

    async def _publish_discovery(self) -> None:
        """Publish the discovery message."""
        jetstream_key_value = await self.broker.key_value(bucket="discovery")

        names = self.provider.names
        for name in names:
            description = self.provider.get_description(name)
            await jetstream_key_value.put(name, description.encode())

    # todo: make it a router
    @asynccontextmanager
    async def lifespan(self, app: Any) -> AsyncIterator[None]:
        async with self.broker:
            await self.broker.start()
            await self._publish_discovery()
            try:
                yield
            finally:
                await self.broker.close()

    def visit_default(self, message: IOMessage) -> None:
        content = message.model_dump()
        logger.debug(f"visit_default(): {content=}")
        syncify(self.broker.publish)(content, self._input_request_subject)

    def visit_text_message(self, message: TextMessage) -> None:
        content = message.model_dump()
        logger.debug(f"visit_text_message(): {content=}")
        syncify(self.broker.publish)(content, self._input_request_subject)

    async def _wait_for_question_response_with_timeout(
        self, question_id: str, *, timeout: int = 180
    ) -> InputResponseModel:
        """Wait for the question response.

        Args:
            question_id (str): The question ID.
            timeout (int, optional): The timeout in seconds. Defaults to 180.
        """
        try:
            # Set a timeout of 180 seconds
            return await asyncio.wait_for(
                self._wait_for_question_response(question_id), timeout=timeout
            )
        except asyncio.TimeoutError:
            logger.debug(
                f"Timeout: User did not send a reply within {timeout} seconds."
            )
            return InputResponseModel(
                msg="User didn't send a reply. Exit the workflow execution.",
                question_uuid=question_id,
                error=True,
            )

    # todo: we need to add timeout and handle it somehow
    async def _wait_for_question_response(self, question_id: str) -> InputResponseModel:
        while True:
            while self.queue.empty():  # noqa: ASYNC110
                await asyncio.sleep(0.1)

            msg: NatsMessage = self.queue.get()
            input_response = InputResponseModel.model_validate_json(
                msg.raw_message.data.decode("utf-8")
            )

            question_id_hex = (
                input_response.question_uuid.hex
                if input_response.question_uuid
                else "None"
            )
            logger.debug(question_id_hex)
            logger.debug(question_id)
            logger.debug(question_id_hex == question_id)
            if question_id_hex == question_id:
                logger.debug("Breaking the while loop")
                break
            else:
                self.queue.put(msg)

        logger.debug("Got the response")
        self.queue.task_done()
        return input_response

    def visit_text_input(self, message: TextInput) -> str:
        content = message.model_dump()
        question_id = message.uuid
        logger.info(f"visit_text_input(): {content=}")
        syncify(self.broker.publish)(content, self._input_request_subject)
        logger.info(
            f"visit_text_input(): published message '{content}' to {self._input_request_subject}"
        )

        input_response: InputResponseModel = syncify(
            self._wait_for_question_response_with_timeout
        )(question_id=question_id)
        logger.info(input_response)
        return input_response.msg

    def visit_multiple_choice(self, message: MultipleChoice) -> str:
        content = message.model_dump()
        question_id = message.uuid
        logger.info(f"visit_multiple_choice(): {content=}")
        syncify(self.broker.publish)(content, self._input_request_subject)

        input_response: InputResponseModel = syncify(
            self._wait_for_question_response_with_timeout
        )(question_id=question_id)
        logger.info(input_response)
        return input_response.msg

    def process_message(self, message: IOMessage) -> Optional[str]:
        try:
            return self.visit(message)
        except Exception as e:
            logger.error(f"Error in process_message: {e}", stack_info=True)
            # do not reraise, we must go on
            if isinstance(message, AskingMessage):
                return "Error: Something went wrong. Please check logs for details."
            return None

    def create_subconversation(self) -> "NatsAdapter":
        return self

    @classmethod
    def create_provider(
        cls,
        nats_url: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
    ) -> ProviderProtocol:
        return NatsProvider(nats_url=nats_url, user=user, password=password)

    @contextmanager
    def create(self, app: Runnable, import_string: str) -> Iterator[None]:
        raise NotImplementedError("NatsAdapter.create() is not implemented")

    def start(
        self,
        *,
        app: Runnable,
        import_string: str,
        name: Optional[str] = None,
        params: dict[str, Any],
        single_run: bool = False,
    ) -> None:
        raise NotImplementedError("NatsAdapter.start() is not implemented")


class NatsProvider(ProviderProtocol):
    def __init__(
        self,
        nats_url: Optional[str] = None,
        user: Optional[str] = None,
        password: Optional[str] = None,
    ) -> None:
        """Initialize the nats workflows.

        Args:
            nats_url (Optional[str], optional): The NATS URL. Defaults to None.
            user (Optional[str], optional): The user. Defaults to None.
            password (Optional[str], optional): The password. Defaults to None.
        """
        self.nats_url = nats_url or "nats://localhost:4222"
        self.user = user
        self.password = password

        self.broker = NatsBroker(self.nats_url, user=self.user, password=self.password)
        self.app = FastStream(self.broker)

        self._initiate_chat_subject: str = "chat.server.initiate_chat"

        self.is_broker_running: bool = False

    async def _setup_subscriber(
        self, ui: UI, from_server_subject: str, to_server_subject: str
    ) -> None:
        logger.info(
            f"Setting up subscriber for {from_server_subject=}, {to_server_subject=}"
        )

        async def consume_msg_from_nats(msg: dict[str, Any], logger: Logger) -> None:
            logger.debug(f"Received message from topic {from_server_subject}: {msg}")
            iomessage = (
                IOMessage.create(**{"type": "error", "long": msg["msg"]})
                if msg.get("error")
                else IOMessage.create(**msg)
            )
            if isinstance(iomessage, AskingMessage):
                processed_message = ui.process_message(iomessage)
                response = InputResponseModel(
                    msg=processed_message, question_uuid=iomessage.uuid
                )
                logger.debug(f"Processed response: {response}")
                await self.broker.publish(response, to_server_subject)
            else:
                ui.process_message(iomessage)

        subscriber = self.broker.subscriber(
            from_server_subject,
            stream=JETSTREAM,
            deliver_policy=api.DeliverPolicy("all"),
        )
        subscriber(consume_msg_from_nats)
        self.broker.setup_subscriber(subscriber)
        await subscriber.start()
        logger.info(f"Subscriber for {from_server_subject} started")

    def run(
        self,
        name: str,
        ui: UI,
        user_id: Optional[str] = None,
        **kwargs: Any,
    ) -> str:
        # subscribe to whatever topic you need
        # consume a message from the topic and call that visitor pattern (which is happening in NatsProvider)
        workflow_uuid = ui._workflow_uuid
        init_message = InitiateWorkflowModel(
            user_id=user_id,
            workflow_uuid=workflow_uuid,
            params=kwargs,
            name=name,
        )
        _from_server_subject = f"chat.client.messages.{user_id}.{workflow_uuid}"
        _to_server_subject = f"chat.server.messages.{user_id}.{workflow_uuid}"

        async def send_initiate_chat_msg() -> None:
            await self.broker.publish(init_message, self._initiate_chat_subject)
            logger.info("Initiate chat message sent")

        @asynccontextmanager
        async def lifespan() -> AsyncIterator[None]:
            async with self.broker:
                await self.broker.start()
                logger.debug("Broker started")
                try:
                    yield
                finally:
                    await self.broker.close()

        async def _setup_and_run() -> None:
            await send_initiate_chat_msg()
            await self._setup_subscriber(ui, _from_server_subject, _to_server_subject)
            while True:  # noqa: ASYNC110
                await asyncio.sleep(0.1)

        async def run_lifespan() -> None:
            if not self.is_broker_running:
                self.is_broker_running = True
                async with lifespan():
                    await _setup_and_run()
            else:
                await _setup_and_run()

        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

        loop.run_until_complete(run_lifespan())

        return "NatsWorkflows.run() completed"

    @asynccontextmanager
    async def _get_jetstream_context(self) -> AsyncIterator[JetStreamContext]:
        nc = NatsClient()
        await nc.connect(self.nats_url, user=self.user, password=self.password)
        js = nc.jetstream()
        try:
            yield js
        finally:
            await nc.close()

    @asynccontextmanager
    async def _get_jetstream_key_value(
        self, bucket: str = "discovery"
    ) -> AsyncIterator[KeyValue]:
        async with self._get_jetstream_context() as js:
            kv = await js.create_key_value(bucket=bucket)
            yield kv

    async def _get_names(self) -> list[str]:
        try:
            async with self._get_jetstream_key_value() as kv:
                names = await kv.keys()
        except NoKeysError:
            names = []
        except NoServersError as e:
            raise FastAgencyNATSConnectionError(
                f"Unable to connect to NATS server at {self.nats_url}"
            ) from e

        return names

    async def _get_description(self, name: str) -> str:
        try:
            async with self._get_jetstream_key_value() as kv:
                description = await kv.get(name)
            return description.value.decode() if description.value else ""
        except KeyNotFoundError as e:
            raise FastAgencyNATSKeyError(
                f"Workflow name {name} not found to get description"
            ) from e
        except NoServersError as e:
            raise FastAgencyNATSConnectionError(
                f"Unable to connect to NATS server at {self.nats_url}"
            ) from e

    @property
    def names(self) -> list[str]:
        names = asyncio.run(self._get_names())
        logger.debug(f"Names: {names}")
        # return ["simple_learning"]
        return names

    def get_description(self, name: str) -> str:
        description = asyncio.run(self._get_description(name))
        logger.debug(f"Description: {description}")
        # return "Student and teacher learning chat"
        return description
